package backend.hdl.arch.rewrite

import backend.hdl.OrderedStreamTypeT
import backend.hdl.arch.{AddInt, AddInt2, ArchLambda, FoldOrderedStream, FoldVector, MapOrderedStream, MapVector, MulInt, OrderedStreamToVector, SplitOrderedStream}
import core.{ParamDef, ParamUse, TypeChecker}
import core.rewrite.{Rule, RulesT}

object ParallelizeDotProductRules extends RulesT {

  override def all(config: Option[Int]): Seq[Rule] =
    Seq(parallelizeDotProduct(config.getOrElse(2))) ++
    Seq(CleanupRules.joinWrites) ++
    RemoveStreamVectorConversionRules.get() ++
    MoveUpVectorToStreamRules.get() ++
    MoveUpJoinStreamRules.all(None) ++
    Seq(MoveUpJoinStreamRules.moveIntoFold) ++
    //      MoveDownStreamToVectorRules.get ++ // try to avoid
    MoveDownSplitStreamRules.get() // try to keep set minimal (use MoveUpJoinStream more)

  def parallelizeDotProduct(parMuls: Int = 2): Rule = Rule("parallelizeDotProduct", {
    case FoldOrderedStream(
    ArchLambda(_, ArchLambda(_, AddInt(_, _), _), _),
    MapOrderedStream(mulfunc @ ArchLambda(_, MulInt(_, _), _), input, mapt: OrderedStreamTypeT), _) if mapt.len.ae.evalInt > 1 && parMuls > 1 =>
      val newInput = TypeChecker.check(SplitOrderedStream(input, parMuls))
      FoldOrderedStream(
        // do not reuse the add function here, because input precision will be different
        AddInt2.asFunction(),
        MapOrderedStream(
          {
            val p = ParamDef(newInput.t.asInstanceOf[OrderedStreamTypeT].et)
            ArchLambda(
              p,
              FoldVector(
                AddInt2.asTypeFunction(),
                MapVector(
                  mulfunc,
                  OrderedStreamToVector(ParamUse(p))
                )
              )
            )
          },
          newInput
        )
      )
  })

}
