package backend.hdl.arch

import backend.hdl.arch.mem.BufferStreamInBlockRam
import backend.hdl._
import core._
import core.util.IRDotGraph
import org.junit.Test

class TypeCheckerTest {

  @Test
  def testZip(): Unit = {
    val tree =
      ZipOrderedStream(Tuple(
        CounterInteger(0, 1, Seq(111)),
        CounterInteger(0, 1, Seq(111))
      ))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == OrderedStreamType(TupleType(IntType(7), IntType(7)), 111))
  }

  @Test
  def testMatrixZip(): Unit = {
    val tree =
      ZipOrderedStream(Tuple(
        Repeat(Repeat(Value(IntType(10)), 111), 222),
        Repeat(Value(IntType(8)), 222)
      ))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == OrderedStreamType(TupleType(OrderedStreamType(IntType(10), 111), IntType(8)), 222))
  }

  @Test
  def testSortingStream(): Unit = {
    val tree =
      UnorderedStreamToOrderedStream(
        Value(UnorderedStreamType(IntType(32), 10))
      )
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == OrderedStreamType(IntType(32), 10))
  }

  @Test
  def testBufferStream(): Unit = {
    val streamType =
      OrderedStreamType(
        UnorderedStreamType(
          OrderedStreamType(IntType(32), 2),
          4
        ),
        10
      )
    val tree = BufferStreamInBlockRam(Value(streamType))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
  }

  @Test
  def testTrunc(): Unit = {
    val tree = MapOrderedStream(TruncInteger.asFunction(Seq(None), Seq(ArithType(8))), Value(OrderedStreamType(IntType(20), 9)))
    val rebuiltTree = TypeChecker.check(tree)
    assert(!rebuiltTree.hasUnknownType)
    assert(rebuiltTree.t == OrderedStreamType(IntType(8), 9))
  }

}
